/*
 *  ftnq.c -- BSO methods implementation
 *
 *  ftnq.c is a part of binkd project
 *
 *  Copyright (C) 1996  Dima Maloff, 5047/13
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. See COPYING.
 */

/*
 * $Id: ftnq.c,v 1.5 1996/12/29 09:42:40 mff Exp mff $
 *
 * $Log: ftnq.c,v $
 * Revision 1.5  1996/12/29  09:42:40  mff
 * Misc changes
 *
 * Revision 1.4  1996/12/14  07:04:21  mff
 * Addedd q_scan_addrs()
 *
 * Revision 1.4  1996/12/04  03:53:55  mff
 * The stable one
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>

#include "Config.h"
#include "ftnq.h"
#include "readcfg.h"
#include "tools.h"
#include "readdir.h"

const char prio[] = "IiCcDdOoFfHh";
static const char flo_flvrs[] = "icdfhICDFH";
static const char out_flvrs[] = "icdohICDOH";

static FTNQ *q_add_dir (FTNQ *q, char *dir, FTN_ADDR *fa1);
static FTNQ *q_add_file (FTNQ *q, char *filename, FTN_ADDR *fa1, char flvr, char action, char type);

/*
 * Frees memory allocated by q_scan()
 */
void q_free (FTNQ *q)
{
  FTNQ *last = 0;

  while (q)
  {
    last = q;
    q = q->next;
  }
  while (last)
  {
    if (!last->prev)
    {
      free (last);
      break;
    }
    else
    {
      last = last->prev;
      if (last->next)
	free (last->next);
    }
  }
}

/*
 * Scans outbound. Return value must be q_free()'d.
 */
FTNQ *q_scan ()
{
  FTNQ *q = 0;				    /* Queue */
  char *s;
  char buf[MAXPATHLEN + 1], outb_path[MAXPATHLEN + 1];

  extern FTN_DOMAIN *pDomains;
  FTN_DOMAIN *curr_domain;

  for (curr_domain = pDomains; curr_domain; curr_domain = curr_domain->next)
  {
    DIR *dp;
    struct dirent *de;
    int len;

    if (curr_domain->alias4 == 0)
    {
      strnzcpy (outb_path, curr_domain->path, sizeof (buf) - 1);
      if (outb_path[strlen (outb_path) - 1] == ':')
	strcat (outb_path, PATH_SEPARATOR);

      if ((dp = opendir (outb_path)) == 0)
      {
	Log (1, "cannot opendir %s: %s", outb_path, strerror (errno));
	continue;
      }

      len = strlen (curr_domain->dir);
      strnzcpy (buf, curr_domain->path, sizeof (buf));
      strnzcpy (buf + strlen (buf), PATH_SEPARATOR, sizeof (buf) - strlen (buf));
      s = buf + strlen (buf);

      while ((de = readdir (dp)) != 0)
      {
	if (!strnicmp (de->d_name, curr_domain->dir, len) &&
	    (de->d_name[len] == 0 ||
	     (de->d_name[len] == '.' && isxdigit (de->d_name[len + 1]))))
	{
	  FTN_ADDR fa;

	  FA_ZERO (&fa);
	  fa.z = ((de->d_name[len] == '.') ?
		  strtol (de->d_name + len + 1, (char **) NULL, 16) :
		  curr_domain->z[0]);
	  strcpy (fa.domain, curr_domain->name);
	  strnzcpy (buf + strlen (buf), de->d_name, sizeof (buf) - strlen (buf));
	  q = q_add_dir (q, buf, &fa);
	  *s = 0;
	}
      }
      closedir (dp);
    }
  }
  return q;
}

/*
 * Adds to the q all files for n akas stored in fa
 */
FTNQ *q_scan_addrs (FTNQ *q, FTN_ADDR *fa, int n)
{
  char buf[MAXPATHLEN + 1];
  int i;
  char *s;

  for (i = 0; i < n; ++i)
  {
    ftnaddress_to_filename (buf, fa + i);
    if (*buf)
    {
      if ((s = max (strrchr (buf, '\\'), strrchr (buf, '/'))) != 0)
      {
	*s = 0;
	q = q_add_dir (q, buf, fa + i);
      }
    }
  }
  return q;
}

/*
 * Adds files from outbound directory _dir_ to _q_. _fa1_ is
 * the address wildcard for this outbound. E.g.
 *     c:\bbs\outbound\                     fa1 = 2:-1/-1.-1@fidonet
 * or
 *     c:\bbs\outbound\00030004.pnt\        fa1 = 2:3/4.-1@fidonet
 * or even
 *     c:\bbs\outbound\00030004.pnt\        fa1 = 2:3/4.5@fidonet
 */
static FTNQ *q_add_dir (FTNQ *q, char *dir, FTN_ADDR *fa1)
{
  DIR *dp;
  FTN_ADDR fa2;
  char buf[MAXPATHLEN + 1];
  int j;
  char *s;

  if ((dp = opendir (dir)) != 0)
  {
    struct dirent *de;

    while ((de = readdir (dp)) != 0)
    {
      s = de->d_name;

      for (j = 0; j < 8; ++j)
	if (!isxdigit (s[j]))
	  break;

      if (j == 8 && strchr (s, 0) - s == 12 &&
	  strchr (s, '.') - s == 8 && strrchr (s, '.') - s == 8)
      {
	/* fa2 will store dest.address for the current (de->d_name) file */
	memcpy (&fa2, fa1, sizeof (FTN_ADDR));
	if (fa1->node != -1 && fa1->p != 0)
	  sscanf (s, "%8x", &fa2.p);	    /* We now in /xxxxyyyy.pnt */
	else
	  sscanf (s, "%4x%4x", &fa2.net, &fa2.node);

	/* add the file if wildcard (f1) match the address (fa2) */
	if (fa1->node == -1 || fa1->p == -1 || !ftnaddress_cmp (fa1, &fa2))
	{
	  strnzcpy (buf, dir, sizeof (buf));
	  strnzcpy (buf + strlen (buf), PATH_SEPARATOR,
		    sizeof (buf) - strlen (buf));
	  strnzcpy (buf + strlen (buf), s, sizeof (buf) - strlen (buf));

	  if (!stricmp (s + 9, "pnt") && fa2.p == -1)
	  {
	    struct stat sb;

	    if (stat (buf, &sb) == 0 && sb.st_mode & S_IFDIR)
	      q = q_add_dir (q, buf, &fa2);
	  }
	  else
	  {
	    if (fa2.p == -1)
	      fa2.p = 0;

	    if (get_node_info (&fa2) || is5D (fa1))
	    {
	      if (strchr (out_flvrs, s[9]) &&
		  tolower (s[10]) == 'u' && tolower (s[11]) == 't')
	      {
		/* Adding *.?ut */
		q = q_add_file (q, buf, &fa2, s[9], 'd', 'm');
	      }
	      else if (strchr (flo_flvrs, s[9]) &&
		       tolower (s[10]) == 'l' && tolower (s[11]) == 'o')
	      {
		/* Adding *.?lo */
		q = q_add_file (q, buf, &fa2, s[9], 'd', 'l');
	      }
	    }
	  }
	}
      }
    }
    closedir (dp);
  }
  else
    Log (1, "cannot opendir %s: %s", dir, strerror (errno));
  return q;
}

/*
 * Add a file to the queue.
 */
static FTNQ *q_add_file (FTNQ *q, char *filename,
			  FTN_ADDR *fa1, char flvr, char action, char type)
{
  FTNQ *new_file = malloc (sizeof (FTNQ));

  assert (new_file != 0);

  FQ_ZERO (new_file);
  new_file->next = q;
  if (q)
    q->prev = new_file;
  q = new_file;

  memcpy (&q->fa, fa1, sizeof (FTN_ADDR));
  strnzcpy (q->path, filename, MAXPATHLEN);
  q->flvr = flvr;
  q->action = action;
  q->type = type;
  q->sent = 0;
  return q;
}

/*
 * Just lists q, not more
 */
void q_list (FILE *out, FTNQ *q)
{
  char buf[FTN_ADDR_SZ + 1];

  fprintf (out, "----\n");
  for (; q; q = q->next)
  {
    if (!q->sent)
    {
      ftnaddress_to_str (buf, &q->fa);
      fprintf (out, "%-20s %c%c%c %8lu %s\n",
	       buf, q->flvr, q->action ? q->action : '-',
	       q->type ? q->type : '-',
	       q->size, q->path);
    }
  }
  fprintf (out, "----\n");
}

/*
 * Selects from q the next file for fa (Returns a pointer to a q element)
 */
FTNQ *select_next_file (FTNQ *q, FTN_ADDR *fa, int nAka)
{
  int j, k;
  FTNQ *curr;

  for (k = 0; k < nAka; ++k)
  {
    for (j = 0; prio[j]; ++j)
    {
      for (curr = q; curr; curr = curr->next)	/* Netmail first */
      {
	if (!curr->sent &&
	    !ftnaddress_cmp (&curr->fa, fa + k) &&
	    curr->flvr == prio[j] &&
	    curr->type == 'm')
	{
	  curr->sent = 1;
	  return curr;
	}
      }
    }
    for (j = 0; prio[j]; ++j)
    {
      for (curr = q; curr; curr = curr->next)	/* Then go files and .flo */
      {
	if (!curr->sent &&
	    !ftnaddress_cmp (&curr->fa, fa + k) &&
	    curr->flvr == prio[j])
	{
	  curr->sent = 1;
	  return curr;
	}
      }
    }
  }
  return 0;
}

FTN_NODE *q_next_node (FTNQ *q)
{
  int i;
  FTNQ *curr;
  FTN_NODE *node;

  for (i = 0; prio[i]; ++i)
  {
    for (curr = q; curr; curr = curr->next)
    {
      /* Make "calls" on flavour-based order. Ignore hold */
      if (!curr->sent &&
	  curr->flvr == prio[i] &&
	  tolower (curr->flvr) != 'h' &&
	  (node = get_node_info (&curr->fa)) != 0)
      {
	/* Found a node to make an outbound call */
	if (strcmp (node->host, "-"))
	{
	  /* Now remove it from the queue */
	  for (curr = q; curr; curr = curr->next)
	  {
	    if (!ftnaddress_cmp (&node->fa, &curr->fa))
	      curr->sent = 1;
	  }
	  return node;
	}
      }
    }
  }
  return 0;
}

/*
 * Creates an empty .?lo
 */
int create_poll (FTN_ADDR *fa, int flvr)
{
  char buf[MAXPATHLEN + 1];
  char ext[5];
  int rc;

  strcpy (ext, ".flo");
  if (flvr)
    ext[1] = flvr;
  ftnaddress_to_filename (buf, fa);
  if (*buf)
  {
    strnzcat (buf, ext, sizeof (buf));
    if ((rc = create_empty_sem_file (buf)) == 0)
      Log (1, "cannot create %s", strerror (errno));
  }
  return rc;
}
