/* $Id: readcfg.c,v 1.3 1996/12/29 09:45:40 mff Exp mff $ */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "readcfg.h"
#include "tools.h"
#include "ftnaddr.h"

const char wsep[] = " \t\n\r\f";

static char *path;
static int line;

int iport = DEF_PORT;
int oport = DEF_PORT;
int call_delay = 60;
int rescan_delay = 60;
int nettimeout = DEF_TIMEOUT;
int oblksize = DEF_BLKSIZE;
int max_servers = 100;
int max_clients = 100;
int kill_dup_partial_files = 0;
int kill_old_partial_files = 0;
int percents = 0;

char sysname[MAXSYSTEMNAME + 1] = "";
char sysop[MAXSYSOPNAME + 1] = "";
char location[MAXLOCATIONNAME + 1] = "";
char nodeinfo[MAXNODEINFO + 1] = "";

char inbound[MAXPATHLEN + 1] = ".";
char inbound_nonsecure[MAXPATHLEN + 1] = "";
char logpath[MAXPATHLEN + 1] = "";
int loglevel = 4;

int nAddr = 0;
FTN_ADDR *pAddr = 0;

extern FTN_DOMAIN *pDomains;

void debug_readcfg ()
{
  int i;
  char buf[80];
  FTN_DOMAIN *curr_domain;

  printf ("addr:");
  for (i = 0; i < nAddr; ++i)
  {
    ftnaddress_to_str (buf, pAddr + i);
    printf (" %s", buf);
  }
  printf ("\n");
  for (curr_domain = pDomains; curr_domain; curr_domain = curr_domain->next)
  {
    if (curr_domain->alias4 == 0)
      printf ("`%s`, `%s`, `%s`.\n",
	      curr_domain->name,
	      curr_domain->path,
	      curr_domain->dir);
    else
      printf ("`%s` alias for `%s`.\n",
	      curr_domain->name,
	      curr_domain->alias4->name);
  }
  printf ("\n");
  for (i = 0; i < nNod; ++i)
  {
    char b[FTN_ADDR_SZ + 1];

    ftnaddress_to_str (b, &pNod[i].fa);
    if (pNod[i].port == DEF_PORT)
      printf ("%s %s %s %s\n", b, pNod[i].host, pNod[i].pwd, pNod[i].filebox);
    else
      printf ("%s %s:%u %s %s\n", b, pNod[i].host, (unsigned) pNod[i].port,
	      pNod[i].pwd, pNod[i].filebox);
  }
  printf ("\n");
}

typedef struct _KEYWORD KEYWORD;
struct _KEYWORD
{
  const char *key;
  void (*callback) (KEYWORD *key, char *s);
  void *var;
  int option1;
  int option2;
};

static void read_aka_list (KEYWORD *, char *);
static void read_domain_info (KEYWORD *, char *);
static void read_node_info (KEYWORD *, char *);
static void read_int (KEYWORD *, char *);
static void read_string (KEYWORD *, char *);
static void read_bool (KEYWORD *, char *);

KEYWORD keywords[] =
{
  {"log", read_string, logpath, 'f', 0},
  {"loglevel", read_int, &loglevel, 0, -1},
  {"domain", read_domain_info, NULL, 0, 0},
  {"address", read_aka_list, NULL, 0, 0},
  {"sysname", read_string, sysname, 0, MAXSYSTEMNAME},
  {"sysop", read_string, sysop, 0, MAXSYSOPNAME},
  {"location", read_string, location, 0, MAXLOCATIONNAME},
  {"nodeinfo", read_string, nodeinfo, 0, MAXNODEINFO},
  {"iport", read_int, &iport, 0, -1},
  {"oport", read_int, &oport, 0, -1},
  {"rescan-delay", read_int, &rescan_delay, 1, -1},
  {"call-delay", read_int, &call_delay, 1, -1},
  {"timeout", read_int, &nettimeout, 1, -1},
  {"oblksize", read_int, &oblksize, MIN_BLKSIZE, MAX_BLKSIZE},
  {"maxservers", read_int, &max_servers, 0, -1},
  {"maxclients", read_int, &max_clients, 0, -1},
  {"inbound", read_string, inbound, 'd', 0},
  {"inbound-nonsecure", read_string, inbound_nonsecure, 'd', 0},
  {"node", read_node_info, NULL, 0, 0},
  {"kill-dup-partial-files", read_bool, &kill_dup_partial_files, 0, 0},
  {"kill-old-partial-files", read_int, &kill_old_partial_files, 1, -1},
  {"percents", read_bool, &percents, 0, 0},
  {NULL, NULL, NULL, 0, 0}
};

#define TEST(var) if (!*var) Log (0, "%s: "#var" should be defined", path)

/*
 * Parses and reads _path as config.file
 */
void readcfg (char *_path)
{
  FILE *in;
  char buf[MAXCFGLINE + 1];
  int i;

  line = 0;
  path = _path;

  if ((in = fopen (path, "r")) == 0)
    Log (0, "%s: %s", path, strerror (errno));

  while (!feof (in))
  {
    if (!fgets (buf, sizeof (buf), in))
      break;
    ++line;

    i = 0;
    while (buf[i] && isspace (buf[i]))
      ++i;
    if (buf[i] && buf[i] != '#')
    {
      char *s = strtok (buf + i, wsep);
      int j;

      if (s)
      {
	for (j = 0; keywords[j].key; ++j)
	  if (!stricmp (keywords[j].key, s))
	    break;

	if (keywords[j].key)
	{
	  keywords[j].callback (keywords + j, strchr (s, 0) + 1);
	}
	else
	{
	  Log (0, "%s: %i: %s: unknown keyword", path, line, s);
	}
      }
    }
  }
  fclose (in);

  TEST (sysname);
  strwipe (sysname);
  TEST (sysop);
  strwipe (sysop);
  TEST (location);
  strwipe (location);
  TEST (nodeinfo);
  strwipe (nodeinfo);

  if (!*inbound_nonsecure)
    strcpy (inbound_nonsecure, inbound);

  if (!nAddr)
    Log (0, "%s: your address should be defined", path);

  if (pDomains == 0)
    Log (0, "%s: at least one domain should be defined", path);

  qsort (pNod, nNod, sizeof (FTN_NODE), (int (*) (const void *, const void *)) node_cmp);

  debug_readcfg ();
}

/*
 *  METHODS TO PROCESS KEYWORDS' ARGUMETS
 */

static void read_aka_list (KEYWORD *key, char *s)
{
  int i;
  char *w;

  for (i = 1; (w = getword (s, i)) != 0; ++i)
  {
    pAddr = realloc (pAddr, sizeof (FTN_ADDR) * (nAddr + 1));
    assert (pAddr != 0);
    if (!parse_ftnaddress (w, pAddr + nAddr))
    {
      Log (0, "%s: %i: %s: the address cannot be parsed", path, line, w);
    }
    if (!is4D (pAddr + nAddr))
    {
      Log (0, "%s: %i: %s: must be at least a 4D address", path, line, w);
    }
    if (!pAddr[nAddr].domain[0])
    {
      if (!pDomains)
	Log (0, "%s: %i: at least one domain must be defined first", path, line);
      strcpy (pAddr[nAddr].domain, get_def_domain ()->name);
    }
    ++nAddr;
    free (w);
  }
}

static void read_domain_info (KEYWORD *key, char *s)
{
  char *w1 = getword (s, 1);
  char *w2 = getword (s, 2);
  char *w3 = getword (s, 3);
  FTN_DOMAIN *new_domain;

  if (!w1 || !w2 || !w3)
    Log (0, "%s: %i: domain: not enough args", path, line);

  if (get_domain_info (w1) == 0)
  {
    new_domain = malloc (sizeof (FTN_DOMAIN));
    assert (new_domain != 0);

    strnzcpy (new_domain->name, w1, sizeof (new_domain->name));

    if (!stricmp (w2, "alias-for"))
    {
      FTN_DOMAIN *tmp_domain;

      if ((tmp_domain = get_domain_info (w3)) == 0)
	Log (0, "%s: %i: %s: undefined domain", path, line, w3);
      else
	new_domain->alias4 = tmp_domain;
      free (w2);
    }
    else
    {
      char *s;
      int z;

      if ((z = atoi (w3)) <= 0)
	Log (0, "%s: %i: invalid zone", path, line);

      new_domain->z = malloc (sizeof (int) * 2);

      assert (new_domain->z != 0);
      new_domain->z[0] = z;
      new_domain->z[1] = 0;
      new_domain->alias4 = 0;

      for (s = w2 + strlen (w2) - 1; (*s == '/' || *s == '\\') && s >= w2; --s)
	*s = 0;
      if ((s = max (strrchr (w2, '\\'), strrchr (w2, '/'))) == 0)
      {
	new_domain->dir = w2;
	new_domain->path = strdup (".");
      }
      else
      {
	new_domain->dir = strdup (s + 1);
	for (; *s == '/' || *s == '\\'; --s)
	  *s = 0;
	new_domain->path = w2;
      }
      if (strchr (new_domain->dir, '.'))
	Log (0, "%s: %i: there should be no extension for "
	     "the base outbound name", path, line);
    }
    new_domain->next = pDomains;
    pDomains = new_domain;
  }
  else
  {
    Log (0, "%s: %i: %s: duplicate domain", path, line, w1);
  }
  free (w1);
  free (w3);
}

static void read_node_info (KEYWORD *key, char *s)
{
  char *w[4];
  int i;

  for (i = 0; i < 4; ++i)
    w[i] = getword (s, i + 1);

  if (getword (s, 5) || !w[0] || !w[1] || !w[2])
    Log (0, "%s: %i: the syntax is incorrect", path, line);

  pNod = realloc (pNod, sizeof (FTN_NODE) * (nNod + 1));

  assert (pNod != 0);

  if (!parse_ftnaddress (w[0], &pNod[nNod].fa))
    Log (0, "%s: %i: %s: the address cannot be parsed", path, line, w[0]);
  else
    exp_ftnaddress (&pNod[nNod].fa);
  if (get_node_info (&pNod[nNod].fa))
    Log (0, "%s: %i: %s: duplicate node entry", path, line, w[0]);

  if ((s = strrchr (w[1], ':')) != 0)
  {
    *s = 0;
    pNod[nNod].port = (unsigned short) atoi (s + 1);
  }
  else
    pNod[nNod].port = (unsigned short) oport;

  strnzcpy (pNod[nNod].host, w[1], sizeof (pNod[nNod].host));
  strnzcpy (pNod[nNod].pwd, w[2], sizeof (pNod[nNod].pwd));
  if (w[3])
    strnzcpy (pNod[nNod].filebox, w[3], sizeof (pNod[nNod].filebox));
  else
    pNod[nNod].filebox[0] = 0;

  ++nNod;

  for (i = 0; i < 4; ++i)
    if (w[i])
      free (w[i]);
}

/*
 * Read a string (key->option1 == 0)
 * or a directory name (key->option1 == 'd')
 * or a file name (key->option1 == 'f')
 */
static void read_string (KEYWORD *key, char *s)
{
  struct stat sb;
  char *target = (char *) (key->var);
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (target, w, key->option1 == 0 ? key->option2 : MAXPATHLEN);
    free (w);

    if (key->option1 != 0)
    {
      w = target + strlen (target) - 1;
      while (w >= target && (*w == '/' || *w == '\\'))
      {
	if (key->option1 == 'f')
	{
	  Log (0, "%s: %i: unexpected `%c' at the end of filename",
	       path, line, *w);
	}
	*(w--) = 0;
      }
      if (key->option1 == 'd' && (stat (target, &sb) == -1 ||
				  !(sb.st_mode & S_IFDIR)))
      {
	Log (0, "%s: %i: %s: incorrect directory", path, line, target);
      }
    }
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void read_int (KEYWORD *key, char *s)
{
  int *target = (int *) (key->var);
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    if ((*target = atoi (w)) < key->option1 ||
	(key->option1 < key->option2 && key->option2 < *target))
      Log (0, "%s: %i: %i: incorrect value", path, line, oblksize);
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void read_bool (KEYWORD *key, char *s)
{
  if (getword (s, 1) == 0)
  {
    *(int *) (key->var) = 1;
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}
